"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const fary_log_1 = require("fary-log");
const koaHttpProxy = require("koa-better-http-proxy");
const env_1 = require("../utils/env");
const is_domain_1 = require("../utils/is-domain");
const server_config_1 = require("../utils/server-config");
const logger = fary_log_1.default(('fary:proxy'));
/**
 * load proxy config
 * @return {*} proxy result
 */
function getProxyConfig() {
    if (!env_1.IS_DEV) {
        return server_config_1.default.proxy;
    }
    // in development mode, clear the cache every time when invoked to improve developing experience!
    delete require.cache[server_config_1.serverConfigPath];
    const newServerConfig = require(server_config_1.serverConfigPath);
    logger.log('Got newest proxy mapping:', newServerConfig.proxy);
    return newServerConfig.proxy;
}
/**
 * proxy middleware
 * @return {Function} koa middleware
 */
function default_1() {
    return function ProxyMiddleware(ctx, next) {
        return __awaiter(this, void 0, void 0, function* () {
            const url = ctx.url;
            let proxyTarget = '';
            let proxyConfig = getProxyConfig();
            // only when non *real production environment*, allow the proxy debugging
            if (!env_1.IS_SERVER_PRODUCTION) {
                // if the `proxyconfig` header is present, then merge it to `proxyConfig` for debugging
                const headerProxyConfigStr = ctx.headers.proxyconfig;
                if (headerProxyConfigStr) {
                    let headerProxyConfig;
                    logger.log(`'proxyconfig' detected in headers: '${headerProxyConfigStr}'`);
                    try {
                        headerProxyConfig = JSON.parse(headerProxyConfigStr);
                    }
                    catch (e) {
                        logger.warn('proxyconfig parsing FAILED, please check it.');
                        headerProxyConfig = {};
                    }
                    proxyConfig = Object.assign({}, proxyConfig, headerProxyConfig);
                    logger.log('Merged proxy config is:', proxyConfig);
                }
            }
            // find the one which match the prefix of `url`
            for (const prefix of Object.keys(proxyConfig)) {
                if (url.startsWith(prefix)) {
                    const target = proxyConfig[prefix];
                    // avoid the multiple proxy problem when local development
                    if (!env_1.IS_DEV || !is_domain_1.default(target)) {
                        ctx.url = url.replace(prefix, '');
                    }
                    proxyTarget = target;
                    logger.log(`Proxy Found: '${prefix}' => '${proxyTarget}'`);
                    break;
                }
            }
            if (!proxyTarget) {
                logger.warn('Proxy NOT FOUND, skipped');
                return next();
            }
            logger.log(`Request '${url}' will be proxied to '${proxyTarget + ctx.url}'`);
            return koaHttpProxy(proxyTarget, {
                // do not parse body, do not limit body size
                parseReqBody: false,
                /**
                 * callback before proxy request
                 * @param {Object} proxyReqOpts - proxy request options
                 * @param {ctx} ctx - koa ctx
                 * @return {Promise.<*>} *
                 */
                proxyReqOptDecorator(proxyReqOpts, context) {
                    return __awaiter(this, void 0, void 0, function* () {
                        // remove header `referer` and `origin` to avoid CSRF interception problem by backend when development
                        if (env_1.IS_DEV) {
                            delete proxyReqOpts.headers.referer;
                            delete proxyReqOpts.headers.origin;
                        }
                        // record the timestamp for proxy duration
                        context.state.proxyStartTimestamp = Date.now();
                        return proxyReqOpts;
                    });
                },
                /**
                 * callback after proxy response
                 * @param {Response} proxyRes - proxy response
                 * @param {Object} proxyResData - response data
                 * @param {ctx} ctx - koa ctx
                 * @return {Promise.<*>} *
                 */
                userResDecorator(proxyRes, proxyResData, context) {
                    return __awaiter(this, void 0, void 0, function* () {
                        logger.info('ProxyRes headers:', proxyRes.headers);
                        const location = `${proxyTarget}${context.url}`;
                        // tslint:disable-next-line
                        logger.log(`Proxy request '${location}' completed(${proxyRes.statusCode}), costing ${Date.now() - context.state.proxyStartTimestamp}ms.`);
                        return proxyResData;
                    });
                },
            })(ctx, next);
        });
    };
}
exports.default = default_1;
